import os
import boto3
import traceback
from . import utils


# submits bucket
SUBMITS_BUCKET = 'autograders-submits'

# autograder files
AUTOGRADERS_BUCKET = 'autograders-defs'

# logs bucket
LOGS_BUCKET = 'autograders-logs'


class DownloadException(Exception):
    '''AWS S3 download exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not download file from AWS S3 Bucket'
        self.traceback = traceback.format_exc()


class UploadException(Exception):
    '''AWS S3 upload exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not upload file to AWS S3 Bucket'
        self.traceback = traceback.format_exc()


def download_submit(uid, submit_id):
    '''Download files to handle submit from AWS S3 Bucket

    Parameters
    ----------
    uid: str
        User identifier
    submit_id: str
        Submit identifier

    Returns
    -------
    None

    Raises
    ------
    DownloadException
        If an error occurs while downloading the files from AWS S3 Bucket
    '''
    try:
        if not os.path.exists('autograders'):
            os.mkdir('autograders')

        if not os.path.exists('autograders/last_modified.pk'):
            utils.write_pickle({}, 'autograders/last_modified.pk')

        last_modified = utils.read_pickle('autograders/last_modified.pk')
        client = boto3.client('s3')
        client.download_file(SUBMITS_BUCKET, f'{submit_id}-{uid}.zip', f'{submit_id}-{uid}.zip')
        utils.extract_to(f'{submit_id}-{uid}.zip', f'{submit_id}-{uid}', delete=True)

        # download autograder if necessary
        info = client.get_object(Bucket=AUTOGRADERS_BUCKET, Key=f'{submit_id}.zip')
        if not os.path.exists(f'{submit_id}.zip') or f'{submit_id}.zip' not in last_modified:
            client.download_file(AUTOGRADERS_BUCKET, f'{submit_id}.zip', f'autograders/{submit_id}.zip')
            last_modified[f'{submit_id}.zip'] = info['LastModified']
            utils.write_pickle(last_modified, 'autograders/last_modified.pk')
        elif last_modified[f'{submit_id}.zip'] < info['LastModified']:
            client.download_file(AUTOGRADERS_BUCKET, f'{submit_id}.zip', f'autograders/{submit_id}.zip')
            last_modified[f'{submit_id}.zip'] = info['LastModified']
            utils.write_pickle(last_modified, 'autograders/last_modified.pk')

        utils.extract_to(f'autograders/{submit_id}.zip', f'{submit_id}-{uid}', delete=False)
    except Exception:
        raise DownloadException()


def upload_submit(uid, submit_id, stream):
    '''Download files to handle submit from AWS S3 Bucket

    Parameters
    ----------
    uid: str
        User identifier
    submit_id: str
        Submit identifier
    stream: stream
        File byte stream

    Returns
    -------
    None

    Raises
    ------
    UploadException
        If an error occurs while uploading the file to AWS S3 Bucket
    '''
    try:
        client = boto3.client('s3')
        client.upload_fileobj(stream, SUBMITS_BUCKET, f'{submit_id}-{uid}.zip')
    except Exception:
        raise UploadException()


def submit_logs(filename):
    '''Tries to upload a log file to the autograders logs AWS S3 Bucket

    Parameters
    ----------
    filename: str
        Filename of the log file to upload

    Returns
    -------
    None
    '''
    try:
        client = boto3.client('s3')
        client.upload_file(filename, LOGS_BUCKET, filename)
    except Exception:
        pass
