import traceback
from . import utils
from . import status
import firebase_admin
from firebase_admin import auth
from firebase_admin import firestore
from firebase_admin import exceptions


class FirebaseException(Exception):
    '''Generic Firebase exception'''

    def __init__(self, message=''):
        super().__init__()
        self.message = message
        self.traceback = traceback.format_exc()


def init():
    '''Initializes Firebase application

    Returns
    -------
    None
    '''
    firebase_admin.initialize_app()


def validate_token(uid):
    '''Validates user identifier

    Parameters
    ----------
    uid: str
        User unique identifier

    Returns
    -------
    bool
        True if the user exists, False if not

    Raises
    ------
    FirebaseException
        If an error occurs while validating user identifier
    '''
    try:
        auth.get_user(uid)
        return True
    except ValueError:
        return False
    except auth.UserNotFoundError:
        return False
    except exceptions.FirebaseError:
        raise FirebaseException('Could not verify user identifier')


def validate_submit_id(submit_id):
    '''Validates submit identifier

    Parameters
    ----------
    submit_id: str
        Submit identifier

    Returns
    -------
    bool
        True if the submit identifier exists, False if not

    Raises
    ------
    FirebaseException
        If an error occurs while validating submit identifier
    '''
    try:
        db = firestore.Client()
        year = str(utils.get_timestamp().year)
        info = db.collection(year).document('ids').get().to_dict() or {}
        return submit_id in info
    except Exception:
        raise FirebaseException('Could not verify submit identifier')


def enqueued(uid, submit_id):
    '''Verifies if a submit is already enqueued

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier

    Returns
    -------
    bool
        True if the submit is already enqueued, False if not

    Raises
    ------
    FirebaseException
        If an error occurs while checking submit enqueued state
    '''
    try:
        db = firestore.Client()
        year = str(utils.get_timestamp().year)
        user_ref = db.collection(year).document('users').collection(uid)
        info = user_ref.document('data').collection('submits').document(submit_id).get().to_dict() or {}
        return info.get('enqueued', False)
    except Exception:
        raise FirebaseException('Could not verify submit enqueued state')


def enqueue_submit(uid, submit_id):
    '''Enqueues a submit

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier

    Returns
    -------
    None

    Raises
    ------
    FirebaseException
        If an error occurs while setting submit in database
    '''
    try:
        db = firestore.Client()
        year = str(utils.get_timestamp().year)

        @firestore.transactional
        def set_submit(transaction):
            user_ref = db.collection(year).document('users').collection(uid)
            total_ref = user_ref.document('total')
            submit_ref = user_ref.document('data').collection('submits').document(submit_id)
            total = total_ref.get(transaction=transaction).to_dict() or {}
            transaction.set(
                total_ref,
                {'value': total.get('value', 0) + 1},
                merge=True
            )
            transaction.set(
                submit_ref,
                {'enqueued': True, 'timestamp': utils.get_timestamp(), 'status': status.PENDING},
                merge=True
            )

        set_submit(db.transaction())
    except Exception:
        raise FirebaseException('Could not enqueue submit')


def set_submit_status(uid, submit_id, stat):
    '''Sets submit status

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier
    stat:
        Submit status

    Returns
    -------
    None

    Raises
    ------
    FirebaseException
        If an error occurs while setting submit status
    '''
    try:
        db = firestore.Client()
        year = str(utils.get_timestamp().year)
        user_ref = db.collection(year).document('users').collection(uid)
        submit_ref = user_ref.document('data').collection('submits').document(submit_id)
        submit_ref.set({'status': stat, 'enqueued': stat in (status.PENDING, status.CHECKING)}, merge=True)
    except Exception:
        raise FirebaseException('Could not set submit status')


def set_submit_result(uid, submit_id, result):
    '''Sets submit result

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier
    result: dict
        Submit result

    Raises
    ------
    FirebaseException
        If an error occurs while setting submit result
    '''
    try:
        db = firestore.Client()
        year = str(utils.get_timestamp().year)
        result['enqueued'] = False
        result['status'] = status.COMPLETED
        user_ref = db.collection(year).document('users').collection(uid)
        submit_ref = user_ref.document('data').collection('submits').document(submit_id)
        submit_ref.set(result, merge=True)
    except Exception:
        raise FirebaseException('Could not set submit result')
