import os
import sys
import shutil
import inspect
import importlib


class CheckException(Exception):
    '''Test suite check exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not run autograder test suite'


class Autograder(object):
    '''Autograders test suite'''

    def setup(self):
        '''Setup method that is called before all defined tests'''
        pass

    def before_each(self):
        '''Method that is called before each defined test'''
        pass

    def after_each(self):
        '''Method that is called after each defined test'''
        pass

    def tear_down(self):
        '''Tear down method that is called after all defined tests'''
        pass

    def check(self, cwd):
        '''Checks all defined test

        Parameters
        ----------
        cwd: str
            Current working directory

        Returns
        -------
        dict
            Submit result
        '''
        notes = {}
        stdouts = ''
        stderrs = ''
        total = 0

        # setup tests
        self.setup()

        # get all _test defined methods
        for (name, method) in inspect.getmembers(self, predicate=inspect.ismethod):
            if '_test' in name:
                self.before_each()
                note, message, stdout, stderr = method(cwd)
                name = method.__doc__ or method.__name__
                total += note
                notes[method.__name__] = {
                    'name': name,
                    'note': note,
                    'message': message
                }
                # append stdout
                if stdout:
                    stdouts += f'{name}:stdout' + (os.linesep * 2)
                    stdouts += stdout + (os.linesep * 2)
                # append stderr
                if stderr:
                    stderrs += f'{name}:stderr' + (os.linesep * 2)
                    stderrs += stderr + (os.linesep * 2)
                self.after_each()

        # tear down test suite
        self.tear_down()

        return {
            'note': total,
            'notes': notes,
            'stdout': stdouts,
            'stderr': stderrs
        }


def run_tests(uid, submit_id):
    '''Run test suite

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier

    Returns
    -------
    None

    Raises
    ------
    CheckException
        If an error occurs while running test suite
    '''
    try:
        submit = f'{submit_id}-{uid}'
        sys.path.append(submit)
        import check
        importlib.reload(check)
        sys.path.remove(submit)

        grader = None
        for n, c in inspect.getmembers(check, predicate=inspect.isclass):
            if Autograder in inspect.getmro(c):
                grader = c()
                result = grader.check(f'./{submit}')
                shutil.rmtree(submit)
                if os.path.exists('__pycache__'):
                    shutil.rmtree('__pycache__')
                return result
    except Exception:
        raise CheckException()
