import os
import pika
import traceback

# RabbitMQ connection
__CONNECTION = None

# RabbitMQ channel
__CHANNEL = None


class ConnectionException(Exception):
    '''RabbitMQ connection exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not connect to the RabbitMQ server'
        self.traceback = traceback.format_exc()


class ConsumeException(Exception):
    '''RabbitMQ consume exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not start consuming submits queue'
        self.traceback = traceback.format_exc()


class PublishException(Exception):
    '''RabbitMQ publish exception'''

    def __init__(self):
        super().__init__()
        self.message = 'Could not publish message to the submits queue'
        self.traceback = traceback.format_exc()


def init():
    '''Sets RabbitMQ connection and channel, declaring submits queue

    Environment Vars
    ----------------
    RABBITMQ_HOST: str (default: localhost)
        RabbitMQ host
    RABBITMQ_USER: str (defualt: guest)
        RabbitMQ username
    RABBITMQ_PASS: str (default: guest)
        RabbitMQ user password

    Returns
    -------
    None
    '''
    global __CONNECTION, __CHANNEL
    rmq_host = os.environ.get('RABBITMQ_HOST', 'localhost')
    rmq_user = os.environ.get('RABBITMQ_USER', 'guest')
    rmq_pass = os.environ.get('RABBITMQ_PASS', 'guest')
    try:
        credentials = pika.PlainCredentials(rmq_user, rmq_pass)
        connection = pika.BlockingConnection(pika.ConnectionParameters(host=rmq_host, credentials=credentials))
        channel = connection.channel()
        channel.queue_declare(queue='submits', durable=True)
        channel.basic_qos(prefetch_count=1)
        __CONNECTION = connection
        __CHANNEL = channel
    except Exception:
        raise ConnectionException()


def close():
    '''Tries to close RabbitMQ connection

    Returns
    -------
    None
    '''
    try:
        __CONNECTION.close()
    except Exception:
        pass


def consume(callback):
    '''Starts consuming submits queue

    Parameters
    ----------
    callback: function
        Callback function to handle a submit message

    Returns
    -------
    None
    '''
    try:
        __CHANNEL.basic_consume(queue='submits', on_message_callback=callback)
        __CHANNEL.start_consuming()
    except Exception:
        raise ConsumeException()


def publish(uid, submit_id):
    '''Publishes a message to the submit queue

    Parameters
    ----------
    uid: str
        User unique identifier
    submit_id: str
        Submit identifier

    Returns
    -------
    None
    '''
    try:
        __CHANNEL.basic_publish(
            exchange='',
            routing_key='submits',
            body=f'{submit_id}-{uid}',
            properties=pika.BasicProperties(delivery_mode=2)
        )
    except Exception:
        raise PublishException()
