import os
import re
import signal
import pickle
import logging
import zipfile
import threading
from subprocess import PIPE
from subprocess import Popen
from datetime import datetime


def get_timestamp():
    '''Gets an UTC timestamp

    Returns
    -------
    datetime
        A datetime object filled with the current UTC date and time
    '''
    return datetime.utcnow()


def get_logger(filename='app.log', file_level='CRITICAL', console_level='ERROR'):
    '''Gets a logger with file and console handlers

    Parameters
    ----------
    filename: str
        Log filename
    file_level: str
        File handler logging level
    console_level: str
        Console handler logging level

    Returns
    -------
    Logger
        Configured logger
    '''
    logger = logging.getLogger('Submits')
    logger.setLevel(logging.__dict__.get(console_level.upper(), 40))
    fh = logging.FileHandler(filename)
    fh.setLevel(logging.__dict__.get(file_level.upper(), 40))
    ch = logging.StreamHandler()
    ch.setLevel(logging.__dict__.get(console_level.upper(), 40))
    formatter = logging.Formatter('%(asctime)s - [%(levelname)s] - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger


def read_file(filename, cwd='.', encoding='utf-8'):
    '''Reads a file

    Parameters
    ----------
    filename: str
        Filename to read
    cwd: str
        Current working directory
    encoding: str
        File encoding

    Returns
    -------
    str
        All the text from the file
    '''
    f = open(os.path.join(cwd, filename), 'r', encoding=encoding)
    text = f.read()
    f.close()
    return text.strip()


def write_pickle(data, filename, cwd='.'):
    '''Writes data to a Pickle file

    Parameters
    ----------
    data: object
        Pickle data to serialize
    filename: str
        Pickle filename
    cwd: str
        Current working directory

    Returns
    -------
    None
    '''
    f = open(os.path.join(cwd, filename), 'wb')
    pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
    f.close()


# reads a json
def read_pickle(filename, cwd='.'):
    '''Reads a Pickle file

    Parameters
    ----------
    filename: str
        Pickle filename to read
    cwd: str
        Current working directory

    Returns
    -------
    object
        Parsed Pickle data from file
    '''
    f = open(os.path.join(cwd, filename), 'rb')
    data = pickle.load(f, encoding="UTF-8")
    f.close()
    return data


# parses a form
def parse_form(filename, cwd='.'):
    '''Parses a form

    Parameters
    ----------
    filename: str
        Form filename
    cwd: str
        Current working directory

    Returns
    -------
    dic
        A dictionary with the answers
    '''
    try:
        f = open(os.path.join(cwd, filename), 'r', encoding='utf-8')
        p = re.compile(r'^[0-9]+( )*:[a-zA-Z0-9, ]+$')
        lookup = {}
        for line in f:
            line = line.strip()
            line = re.sub(r'( |\t)+', ' ', line)
            line = re.sub(r'( |\t)*,( |\t)*', ',', line)
            if p.search(line) is not None:
                key, value = line.split(':')
                lookup[key.strip()] = value.strip()
        return lookup
    except Exception:
        return {}


def extract_to(filename, to, delete=False):
    '''Extracts zip file content to the specified directory

    Parameters
    ----------
    filename: str
        Zip filename
    to: str
        destination directory
    delete: bool
        deletes the extracted zip file

    Returns
    -------
    None
    '''
    zip_ref = zipfile.ZipFile(filename, 'r')
    zip_ref.extractall(to)
    zip_ref.close()
    if delete:
        os.remove(filename)


def execute(cmd=[], cwd='.', shell=False, input=None, encoding='utf-8', timeout=5):
    '''Executes a command

    Parameters
    ----------
    cmd: list or str
        Command to execute
    cwd: str
        Current working directory
    shell: bool
        If shell is True, the specified command will be executed through the shell
    input: str or None
        Input to write in stdin
    encoding: str
        Input encoding
    timeout: int
        Execution timeout in seconds

    Returns
    -------
    ExecuteResult
        Execution result
    '''
    class ExecuteResult:
        def __init__(self):
            self.stdout = ''
            self.stderr = ''
            self.code = 0
            self.timeout = False

        def __str__(self):
            return self.stdout

    class Execute(threading.Thread):

        def __init__(self, cmd, shell, input, encoding, cwd='.'):
            super().__init__()
            self._stop_event = threading.Event()
            self.process = Popen(cmd, shell=shell, cwd=cwd, stdout=PIPE, stdin=PIPE, stderr=PIPE, preexec_fn=os.setsid)
            if input is not None and isinstance(input, str):
                self.process.stdin.write((input + os.linesep).encode(encoding))
                self.process.stdin.flush()
            self.result = ExecuteResult()

        def stop(self):
            self.result.timeout = True
            self._stop_event.set()

        def run(self):
            while not self._stop_event.is_set():
                if self.process.poll() is not None:
                    self.result.stdout = self.process.stdout.read().decode().rstrip()
                    self.result.stderr = self.process.stderr.read().decode().rstrip()
                    self.result.code = self.process.returncode
                    self.result.timeout = False
                    break

            if self.result.timeout:
                os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

    t = Execute(cmd, shell, input, encoding, cwd=cwd)
    t.start()
    t.join(timeout=timeout)
    t.stop()
    return t.result


def make(target='', timeout=5):
    '''Executes a command

    Parameters
    ----------
    target: str
        Makefile target name
    timeout: int
        Execution timeout in seconds

    Returns
    -------
    ExecuteResult
        Execution result
    '''
    return execute(cmd=['make', target], timeout=timeout)
